% Code to solve the eigenvalue problem for the linearized 3-layer
% shallow-water system in cylindrical coordinates.
%
% Uses background fields read from the Aronnax model.

% P. E. Isachsen, April 2019, March 2022, Sep-22

% Warning: Adjusting the grid a bit here. Assume the data comes on
% v/eta points. Then make a staggered grid where v/eta points have
% been shifted half a grid spacing to the left.

f = 1.45e-4;        % Coriolis parameter [s-1]}
rho0 = 1028;        % Background density [kg m-3]
Lr = 750e3;         % Length of the (radial) transect [m]
rb = 2e-6           % Bottom damping coefficient [s-1]
g = 9.81;           % Gravitational acceleration [m s-2]
gp12 = 0.024;       % Reduced gravity layer 1 to 2 [m s-2]
gp23 = 0.008;       % Reduced gravity layer 2 to 3 [m s-2]

dname = 'INSERT_DIRECTORY';

% Radial grid
r = ncread([dname,'/BEAU_VERT.nc'],'r'); r = r(:)';
r = double(r);
Nr = length(r);
% Grid spacing
dr = mean(diff(r));

% 'External' ru points including wall points (where we'll have no
% equations, only boundary conditions u=0)
ruext = 0:dr:Lr;
% internal ru points
ru = ruext(2:end-1);
% r points for v and eta
r = (ruext(1:end-1)+ruext(2:end))/2;

% Number of cells (v/eta points)
N = length(r);
% Number of u points
Nu = N-1;

% Read data from a specific run
%run = 1; fname = [dname,'/BEAU_VERT.nc'];
%run = 2; fname = [dname,'/BEAU_0015.nc'];
%run = 3; fname = [dname,'/BEAU_002.nc'];
run = 4; fname = [dname,'/BEAU_004.nc'];
%run = 5; fname = [dname,'/BEAU_006.nc'];

V = ncread(fname,'azimuthal_velocity_atm');
V1 = V(:,1); V1 = V1(:)'; % Azimuthal velocity in layer 1
V2 = V(:,2); V2 = V2(:)'; % ...layer 2
V3 = V(:,3); V3 = V3(:)'; % ...layer 3
clear V

% Azimuthal velocities on u points
V1u = (V1(1:end-1)+V1(2:end))/2;
V2u = (V2(1:end-1)+V2(2:end))/2;
V3u = (V3(1:end-1)+V3(2:end))/2;

% Background radial velocity gradient (on eta/v points)
% Note: first extrapolate onto external (wall) u points
V1uext = [V1(1) V1u V1(end)];
V2uext = [V2(1) V2u V2(end)];
V3uext = [V3(1) V3u V3(end)];
%
V1r = diff(V1uext)/dr;
V2r = diff(V2uext)/dr;
V3r = diff(V3uext)/dr;

H = ncread(fname,'layer_thickness_atm');
H1 = H(:,1); H1 = H1(:)'; % thickness layer 1
H2 = H(:,2); H2 = H2(:)'; % ...layer 2
H3 = H(:,3); H3 = H3(:)'; % ...layer 3
clear H
Eta1 = ncread(fname,'eta_atm'); Eta1 = Eta1(:)'; % SSH

% Total water column thickness
H = H1+H2+H3;
% Z-levels
zEta2 = -(H1-Eta1);
zEta3 = zEta2-H2;
zb = -(H-Eta1);

% Layer thicknesses on u points
H1u = (H1(1:end-1)+H1(2:end))/2;
H2u = (H2(1:end-1)+H2(2:end))/2;
H3u = (H3(1:end-1)+H3(2:end))/2;

% Azimuthal velocities on u points
V1u = (V1(1:end-1)+V1(2:end))/2;
V2u = (V2(1:end-1)+V2(2:end))/2;
V3u = (V3(1:end-1)+V3(2:end))/2;

% Background radial velocity gradient (on eta/v points)
% Note: first extrapolate onto external (wall) u points
V1uext = [V1(1) V1u V1(end)];
V2uext = [V2(1) V2u V2(end)];
V3uext = [V3(1) V3u V3(end)];
%
V1r = diff(V1uext)/dr;
V2r = diff(V2uext)/dr;
V3r = diff(V3uext)/dr;
%Eta2r = diff(Eta2)/dr;
%Eta3r = diff(Eta3)/dr;


% Set up linearized stability problem

% Starting indeces (minus 1) in matrix for u, v and eta equations of each layer

m0u1 = 0;
m0v1 = m0u1+(N-1);
m0e1 = m0v1+N;
m0u2 = m0e1+N;
m0v2 = m0u2+(N-1);
m0e2 = m0v2+N;
m0u3 = m0e2+N;
m0v3 = m0u3+(N-1);
m0e3 = m0v3+N;

% azimuthal wave numbers
l_wave = 1:1:40;

Nl = length(l_wave);

nmodes = 6; % we'll just keep track of the nmodes
            % fastest-growing modes at each wavenumber

% Storage variables for imaginary and real frequencies
omegaim = zeros(Nl,nmodes); % imaginary frequencies
omegarm = zeros(Nl,nmodes); % real frequencies

% Vorticity fluxes and layer thickness fluxes (form drag)
u1zeta1m = zeros(Nl,nmodes,Nu);
u2zeta2m = zeros(Nl,nmodes,Nu);
u3zeta3m = zeros(Nl,nmodes,Nu);
fd1m = zeros(Nl,nmodes,Nu);
fd2m = zeros(Nl,nmodes,Nu);
fd3m = zeros(Nl,nmodes,Nu);

% Prepare some work variables
NA = (N-1+2*N)+(N-1+2*N)+(N-1+2*N); % Size of coefficient matrix

H1A = repmat(H1,[NA,1]);
H2A = repmat(H2,[NA,1]);
H3A = repmat(H3,[NA,1]);
V1rA = repmat(V1r,[NA,1]);
V2rA = repmat(V2r,[NA,1]);
V3rA = repmat(V3r,[NA,1]);
%Eta2rA = repmat(Eta2r,[NA,1]);
%Eta3rA = repmat(Eta3r,[NA,1]);
V1mV2A = repmat(V1-V2,[NA,1]);
V2mV3A = repmat(V2-V3,[NA,1]);
rA = repmat(r,[NA,1]);

i = sqrt(-1);

% Construct problem and solve for each l wavenumber
for k = 1:Nl

   disp([int2str(k), ' out of ', int2str(Nl), ' wavenumbers.']);

   l = l_wave(k);

   % Prepare the coefficient matrix
   A = zeros(NA,NA);

   % u1 equations
   for n = 1:N-1
      m = m0u1+n;
      A(m,m)        = l*V1u(n)/ru(n);
      A(m,m0v1+n)   = i*(2*V1u(n)/ru(n)+f)/2;
      A(m,m0v1+n+1) = i*(2*V1u(n)/ru(n)+f)/2;
      % (V+v)(V+v) = V^2+2*Vv+v^2
      A(m,m0e1+n)   = i*g/dr;
      A(m,m0e1+n+1) = -i*g/dr;
   end % for n...

   % v1 equations
   for n = 1:N
      m = m0v1+n;
      if n>1
      A(m,m0u1+n-1) = -i*(V1r(n)+f+V1(n)/r(n))/2;
      end
      if n<N
      A(m,m0u1+n)   = -i*(V1r(n)+f+V1(n)/r(n))/2;
      end
      A(m,m)        = l*V1(n)/r(n);
      A(m,m0e1+n)   = l*g/r(n);
   end % for n...

   % eta1 equations    
   for n = 1:N
      m = m0e1+n;
      if n>1
      A(m,m0u1+n-1) = i*H1u(n-1)/dr-i*H1(n)/r(n)/2;
      end
      if n<N
      A(m,m0u1+n)   = -i*H1u(n)/dr-i*H1(n)/r(n)/2;
      end
      A(m,m0v1+n)   = l*H1(n)/r(n);
      A(m,m)        = l*V1(n)/r(n);
      % The following is a substitution for the expression for d(eta2)/dt
      if n>1
      A(m,m0u2+n-1) = i*H2u(n-1)/dr-i*H2(n)/r(n)/2;
      end
      if n<N
      A(m,m0u2+n)   = -i*H2u(n)/dr-i*H2(n)/r(n)/2;
      end
      A(m,m0v2+n)   = l*H2(n)/r(n);
      A(m,m0e2+n)   = l*(V2(n)-V1(n))/r(n);
      if n>1
      A(m,m0u3+n-1) = i*H3u(n-1)/dr-i*H3(n)/r(n)/2;
      end
      if n<N
      A(m,m0u3+n)   = -i*H3u(n)/dr-i*H3(n)/r(n)/2;
      end
      A(m,m0v3+n)   = l*H3(n)/r(n);
      A(m,m0e3+n)   = l*(V3(n)-V2(n))/r(n);
   end % for n...

   % u2 equations
   for n = 1:N-1
      m = m0u2+n;
      A(m,m)        = l*V2u(n)/ru(n);
      A(m,m0v2+n)   = i*(2*V2u(n)/ru(n)+f)/2;
      A(m,m0v2+n+1) = i*(2*V2u(n)/ru(n)+f)/2;
      A(m,m0e1+n)   = i*g/dr;
      A(m,m0e1+n+1) = -i*g/dr;
      A(m,m0e2+n)   = i*gp12/dr;
      A(m,m0e2+n+1) = -i*gp12/dr;
   end % for n...

   % v2 equations
   for n = 1:N
      m = m0v2+n;
      if n>1
      A(m,m0u2+n-1) = -i*(V2r(n)+f+V2(n)/r(n))/2;
      end
      if n<N
      A(m,m0u2+n)   = -i*(V2r(n)+f+V2(n)/r(n))/2;
      end
      A(m,m)        = l*V2(n)/r(n);
      A(m,m0e1+n)   = l*g/r(n);
      A(m,m0e2+n)   = l*gp12/r(n);
   end % for n...

   % eta2 equations    
   for n = 1:N
      m = m0e2+n;
      if n>1
      A(m,m0u2+n-1) = i*H2u(n-1)/dr-i*H2(n)/r(n)/2;
      end
      if n<N
      A(m,m0u2+n)   = -i*H2u(n)/dr-i*H2(n)/r(n)/2;
      end
      A(m,m0v2+n)   = l*H2(n)/r(n);
      A(m,m)        = l*V2(n)/r(n);
      % The following is a substitution for the expression for d(eta3)/dt
      if n>1
      A(m,m0u3+n-1) = i*H3u(n-1)/dr-i*H3(n)/r(n)/2;
      end
      if n<N
      A(m,m0u3+n)   = -i*H3u(n)/dr-i*H3(n)/r(n)/2;
      end
      A(m,m0v3+n)   = l*H3(n)/r(n);
      A(m,m0e3+n)   = l*(V3(n)-V2(n))/r(n);

   end % for n...

   % u3 equations
   for n = 1:N-1
      m = m0u3+n;
      A(m,m)        = l*V3u(n)/ru(n);
      A(m,m0v3+n)   = i*(2*V3u(n)/ru(n)+f)/2;
      A(m,m0v3+n+1) = i*(2*V3u(n)/ru(n)+f)/2;
      A(m,m0e1+n)   = i*g/dr;
      A(m,m0e1+n+1) = -i*g/dr;
      A(m,m0e2+n)   = i*gp12/dr;
      A(m,m0e2+n+1) = -i*gp12/dr;
      A(m,m0e3+n)   = i*gp23/dr;
      A(m,m0e3+n+1) = -i*gp23/dr;
   end % for n...

   % v3 equations
   for n = 1:N
      m = m0v3+n;
      if n>1
      A(m,m0u3+n-1) = -i*(V3r(n)+f+V3(n)/r(n))/2;
      end
      if n<N
      A(m,m0u3+n)   = -i*(V3r(n)+f+V3(n)/r(n))/2;
      end
      A(m,m)        = l*V3(n)/r(n);
      A(m,m0e1+n)   = l*g/r(n);
      A(m,m0e2+n)   = l*gp12/r(n);
      A(m,m0e3+n)   = l*gp23/r(n);
   end % for n...

   % eta3 equations    
   for n = 1:N
      m = m0e3+n;
      if n>1
      A(m,m0u3+n-1) = i*H3u(n-1)/dr-i*H3(n)/r(n)/2;
      end
      if n<N
      A(m,m0u3+n)   = -i*H3u(n)/dr-i*H3(n)/r(n)/2;
      end
      A(m,m0v3+n)   = l*H3(n)/r(n);
      A(m,m)        = l*V3(n)/r(n);
   end % for n...

   % Solve eigenvalue problem
   [eigvec,eigval] = eig(A);

   % Extract eigenvalue and eigenvectors
   omega = diag(eigval);
   u1    = eigvec(m0u1+1:m0u1+(N-1),:);
   v1    = eigvec(m0v1+1:m0v1+N,:);
   eta1  = eigvec(m0e1+1:m0e1+N,:);
   u2    = eigvec(m0u2+1:m0u2+(N-1),:);
   v2    = eigvec(m0v2+1:m0v2+N,:);
   eta2  = eigvec(m0e2+1:m0e2+N,:);
   u3    = eigvec(m0u3+1:m0u3+(N-1),:);
   v3    = eigvec(m0v3+1:m0v3+N,:);
   eta3  = eigvec(m0e3+1:m0e3+N,:);

   % Sort according to fastest-growing modes
   omegai = imag(omega);
   omegar = real(omega);
   [~,kk] = sort(omegai,'descend');
   omegai = omegai(kk);
   omegar = omegar(kk);
   u1 = u1(:,kk);
   v1 = v1(:,kk);
   eta1 = eta1(:,kk);
   u2 = u2(:,kk);
   v2 = v2(:,kk);
   eta2 = eta2(:,kk);
   u3 = u3(:,kk);
   v3 = v3(:,kk);
   eta3 = eta3(:,kk);
   
   % Do (non-conjugate) transpose
   u1 = u1.';
   v1 = v1.';
   eta1 = eta1.';
   u2 = u2.';
   v2 = v2.';
   eta2 = eta2.';
   u3 = u3.';
   v3 = v3.';
   eta3 = eta3.';
   
   % Interpolate u velocities onto v/eta points
   u1uext = cat(2,zeros(NA,1),u1,zeros(NA,1));
   u1v = (u1uext(:,1:end-1)+u1uext(:,2:end))/2;
   u2uext = cat(2,zeros(NA,1),u2,zeros(NA,1));
   u2v = (u2uext(:,1:end-1)+u2uext(:,2:end))/2;
   u3uext = cat(2,zeros(NA,1),u3,zeros(NA,1));
   u3v = (u3uext(:,1:end-1)+u3uext(:,2:end))/2;

   % MKE-to-EKE conversion (for each layer)
   Cbt1 = -H1A.*(u1v.*conj(v1)).*V1rA;
   Cbt2 = -H2A.*(u2v.*conj(v2)).*V2rA;
   Cbt3 = -H3A.*(u3v.*conj(v3)).*V3rA;
   % MAPE-to-EAPE conversion
   Cbc12 = g./rA.*V1mV2A.*(eta1.*conj(i*l*eta2));
   Cbc13 = g./rA.*V2mV3A.*(eta1.*conj(i*l*eta3));
   Cbc23 = gp12./rA.*V2mV3A.*(eta2.*conj(i*l*eta3));

   % Interpolate v velocities onto internal u points
   v1u = (v1(:,1:end-1)+v1(:,2:end))/2;
   v2u = (v2(:,1:end-1)+v2(:,2:end))/2;
   v3u = (v3(:,1:end-1)+v3(:,2:end))/2;

   v1r = diff(v1,1,2)/dr;
   v2r = diff(v2,1,2)/dr;
   v3r = diff(v3,1,2)/dr;

   % Vorticity
   zeta1 = v1u./ru + v1r - i*l*u1./ru;
   zeta2 = v2u./ru + v2r - i*l*u2./ru;
   zeta3 = v3u./ru + v3r - i*l*u3./ru;

   % Vorticity flux
   u1zeta1 = u1.*conj(zeta1);
   u2zeta2 = u2.*conj(zeta2);
   u3zeta3 = u3.*conj(zeta3);

   % Layer thickness flux
   h1 = eta1-eta2;
   h2 = eta2-eta3;
   h3 = eta3;

   h1u = (h1(:,1:end-1)+h1(:,2:end))/2;
   h2u = (h2(:,1:end-1)+h2(:,2:end))/2;
   h3u = (h3(:,1:end-1)+h3(:,2:end))/2;

   u1h1 = u1.*conj(h1u);
   u2h2 = u2.*conj(h2u);
   u3h3 = u3.*conj(h3u);

   % Form drag formulation
   fd1 = f*u1h1./repmat(H1u,[NA 1]);
   fd2 = f*u2h2./repmat(H2u,[NA 1]);
   fd3 = f*u3h3./repmat(H3u,[NA 1]);
   
   % Keep track of frequencies of the fastest-growing modes
   omegaim(k,:) = omegai(1:nmodes)';
   omegarm(k,:) = omegar(1:nmodes)';

   u1zeta1m(k,:,:) = u1zeta1(1:nmodes,:);
   u2zeta2m(k,:,:) = u2zeta2(1:nmodes,:);
   u3zeta3m(k,:,:) = u3zeta3(1:nmodes,:);
   fd1m(k,:,:) = fd1(1:nmodes,:);
   fd2m(k,:,:) = fd2(1:nmodes,:);
   fd3m(k,:,:) = fd3(1:nmodes,:);

end % for k...
